class HexBoard {
	// The dimensions of the board
	public int sizeX, sizeY;
	// The board is a two-dimensional array of integers, 
	// denoting the tile assigned to each location	
	int boardTiles[][];
   int boardOrient[][];

	public Violations violations;
	
	public TileSet Tiles;	

	private Define define;   
	
	int conflict[];
	int numInConflict;
	
	// Constructor: Declares and initialises memory
	public HexBoard(int boardSizeX, int boardSizeY, String tilesToUse) {
		int x, y;

		boardTiles = new int[boardSizeX][boardSizeY];
		boardOrient = new int[boardSizeX][boardSizeY];
		
		conflict = new int[boardSizeX*boardSizeY];
		numInConflict=0;
		
		violations = new Violations(boardSizeX, boardSizeY);
		
		for (x=0; x<boardSizeX; x++) {
			for (y=0; y<boardSizeY; y++) {
				boardTiles[x][y] = -1;
				boardOrient[x][y] = -1;
			}
		}
		
		Tiles = new TileSet(tilesToUse);
		
		sizeX = boardSizeX;
		sizeY = boardSizeY;

	}
	
	// Puts a tile on the board at position <x, y> with orientation orient
	public boolean PlaceTile(int tile, int orient, int x, int y) {
		boardTiles[x][y] = tile;
		boardOrient[x][y] = orient;
		return true;
	}
	
	public int TILE(int x, int y) {
		return (y+x*sizeY);
	}
	
	public void RemoveViolations(int x1, int y1) {
		// Location and tile numbers of two tiles
		int x2, y2, t1, t2;
		// Relevant touching edges of two tiles
		int e1, e2;
		// Colour code of these edges
		char c1, c2;		
	
		int i, v=0;
	
	
		Adjacent a;
		a = GetAdjacent(x1, y1, define.ALL_ADJACENT);
	
		t1 = boardTiles[x1][y1];
		
		// and iterate through them
		for (i=0; i<a.numAdjacent; i++) {
			// Get the number of the edge that is adjacent
			e1 = (6 + a.dir[i] - boardOrient[x1][y1])%6;
			// Find the colour at that edge
			c1 = Tiles.TileDesc[t1].charAt(e1);
	
			// Get the details of the adjacent tile										
			x2 = a.xs[i];
			y2 = a.ys[i];
			t2 = boardTiles[x2][y2];
			
			// And its edge and colour details
			e2 = (6 + (a.dir[i]+3)%6 - boardOrient[x2][y2])%6;
			c2 = Tiles.TileDesc[t2].charAt(e2);
			
			// Check if there is a current violation, and remove it
			if (c1!=c2) {
				violations.byTile[x1][y1]--;
				violations.byTile[x2][y2]--;
				violations.crossViolations[TILE(x1,y1)][TILE(x2,y2)]=0;
				violations.crossViolations[TILE(x2,y2)][TILE(x1,y1)]=0;
				//violations.byTile[x2][y2]++;
				v+=2;
			} 
		}
		// Remove all violations found
		violations.total-=v;
	}
	
	public void UpdateViolations(int x1, int y1) {
		// Location and tile numbers of two tiles
		int x2, y2, t1, t2;
		// Relevant touching edges of two tiles
		int e1, e2;
		// Colour code of these edges
		char c1, c2;		
	
		int i, v=0;
	
	
		Adjacent a;
		a = GetAdjacent(x1, y1, define.ALL_ADJACENT);
	
		t1 = boardTiles[x1][y1];
	
		// and iterate through them
		for (i=0; i<a.numAdjacent; i++) {
			// Get the number of the edge that is adjacent
			e1 = (6 + a.dir[i] - boardOrient[x1][y1])%6;
			// Find the colour at that edge
			c1 = Tiles.TileDesc[t1].charAt(e1);
	
			// Get the details of the adjacent tile										
			x2 = a.xs[i];
			y2 = a.ys[i];
			t2 = boardTiles[x2][y2];
			
			// And its edge and colour details
			e2 = (6 + (a.dir[i]+3)%6 - boardOrient[x2][y2])%6;
			c2 = Tiles.TileDesc[t2].charAt(e2);
			
			// Check if there is a current violation, and remove it
			if (c1!=c2) {
				violations.byTile[x1][y1]++;
				violations.byTile[x2][y2]++;
				violations.crossViolations[TILE(x1,y1)][TILE(x2,y2)]=1;
				violations.crossViolations[TILE(x2,y2)][TILE(x1,y1)]=1;
				//violations.byTile[x2][y2]++;
				v+=2;
			} 
		}
		// Remove all violations found
		violations.total+=v;
	}
	
	// This regenerates this list of tiles in conflict each time it is 
	// called so could be improved
	public void UpdateConflicts() {
		int x1, y1;
	
		numInConflict=0;
		for (x1=0; x1<sizeX; x1++) {
			for (y1=0; y1<sizeY; y1++) {
				if (violations.byTile[x1][y1]>0) {
					conflict[numInConflict++] = TILE(x1,y1);
				}
			}
		}
	}

	// Swaps two tiles, <x1, y1> with <x2, y2>. Tile 1 is placed with new orientation
	// no1 and tile 2 with new orientation no2.
	public void SwapTiles(int x1, int y1, int x2, int y2, int no1, int no2) {
		int t1, t2;

		// Fix for double counting
		if (InViolation(x1, y1, x2, y2)) {
			violations.byTile[x1][y1]++;
			violations.byTile[x2][y2]++;
			violations.total+=2;		
		}
	
		
		RemoveViolations(x1, y1);
		// Handle the case were you a swapping a tile with itself.
		// That is, you are simply changing its orientation
		if (x1!=x2 || y1!=y2) RemoveViolations(x2, y2);
	
		// Get tile numbers
		t1 = boardTiles[x1][y1];
		t2 = boardTiles[x2][y2];
	
		// Swap them
		boardTiles[x1][y1] = t2;
		boardOrient[x1][y1] = no2;
		boardTiles[x2][y2] = t1;
		boardOrient[x2][y2] = no1;
	
		
		// Update the violations for the new stat of the tiles
		UpdateViolations(x1, y1);
		if (x1!=x2 || y1!=y2) UpdateViolations(x2, y2);
	
		// Fix for double counting
		if (InViolation(x1, y1, x2, y2)) {
			violations.byTile[x1][y1]--;
			violations.byTile[x2][y2]--;
			violations.total-=2;		
		}
	
		// Keep track of the tiles in conflict
		UpdateConflicts();
	}

// Returns the number of violations on the current board
	public int CountViolations() {
		// Location and tile numbers of two tiles
		int x1, y1, x2, y2, t1, t2;
		// Relevant touching edges of two tiles
		int e1, e2;
		// Colour code of these edges
		char c1, c2;		

		int i, v=0;
		Adjacent a;
		
		// Initialise the violation matrix for each tile on the board
		for (x1=0; x1<sizeX; x1++) {
			for (y1=0; y1<sizeY; y1++) {
				violations.byTile[x1][y1] = 0;
			}
		}
		
		// Look at every tile on the board
		for (x1=0; x1<sizeX; x1++) {
			for (y1=0; y1<sizeY; y1++) {
				// Get the tile number
				t1 = boardTiles[x1][y1];

				// Get a list of all tiles adjacent to this one
				a = GetAdjacent(x1, y1, define.ALL_ADJACENT);
				// and iterate through them
				for (i=0; i<a.numAdjacent; i++) {
					// Get the number of the edge that is adjacent
					e1 = (6 + a.dir[i] - boardOrient[x1][y1])%6;
					// Find the colour at that edge
					c1 = Tiles.TileDesc[t1].charAt(e1);

					// Get the details of the adjacent tile										
					x2 = a.xs[i];
					y2 = a.ys[i];
					t2 = boardTiles[x2][y2];
					
					// And its edge and colour details
					e2 = (6 + (a.dir[i]+3)%6 - boardOrient[x2][y2])%6;
					c2 = Tiles.TileDesc[t2].charAt(e2);
					
					// Check if there is a violation, and if so count it
					if (c1!=c2) {
						violations.byTile[x1][y1]++;
						violations.crossViolations[TILE(x1,y1)][TILE(x2,y2)]=1;
						v++;
					} else {
						violations.crossViolations[TILE(x1,y1)][TILE(x2,y2)]=0;				
					}

				}
			}
		}
		violations.total = v;
      UpdateConflicts();
		return v;
	}

	// Returns boolean true iff two tiles are in violation
	public boolean InViolation(int x1, int y1, int x2, int y2) {
		int t1, t2, e1, e2;
		char c1, c2;
		int i;
		Adjacent a;
		a = GetAdjacent(x1, y1, define.ALL_ADJACENT);
		
		// Get tile numbers
		t1 = boardTiles[x1][y1];
		t2 = boardTiles[x2][y2];
		
		for (i=0; i<a.numAdjacent; i++) {
			if ((a.xs[i]==x2) && (a.ys[i]==y2)) {
				// Get the number of the edge that is adjacent
				e1 = (6 + a.dir[i] - boardOrient[x1][y1])%6;
				// Find the colour at that edge
				c1 = Tiles.TileDesc[t1].charAt(e1);

				// And corresponding edge and colour details
				e2 = (6 + (a.dir[i]+3)%6 - boardOrient[x2][y2])%6;
				c2 = Tiles.TileDesc[t2].charAt(e2);
					
				if (c1!=c2) {
					return true;
				} else {
					return false;
				}
			}
		}
		
		return false;
	}


	// Returns boolean true iff two locations are adjacent
	public boolean IsAdjacent(int x1, int y1, int x2, int y2) {
		int i;
		Adjacent a;
		a = GetAdjacent(x1, y1, define.ALL_ADJACENT);
		
		for (i=0; i<a.numAdjacent; i++) {
			if ((a.xs[i]==x2) && (a.ys[i]==y2)) return true;
		}
		return false;
	}

// This function returns tiles adjacent to another tile
// which is a constant and may be either ALL_ADJACENT to get
// all tiles or FWD_ADJACENT which is only the first three directions
	public Adjacent GetAdjacent(int x, int y, int which) {
		int d, max;
		
		int xa, ya;
		
		Adjacent a = new Adjacent();
		
		// Select which adjacent tiles we are interested in
		if (which==define.FWD_ADJACENT) {
			max = 3;
		} else {
			max = 6;
		}		

		// Enumerate through the adjacent tiles we are interested in
		// selecting only those within the confines of the board
		for (d=0; d<max; d++) {
			switch (d) {
				case 0:	// UP
					xa = x;
					ya = y - 1;
					break;
				
				case 1:	// UP-LEFT
					xa = x+1;
					ya = y - 1 + x % 2;
					break;

				case 2:	// DOWN-LEFT
					xa = x+1;
					ya = y + x % 2;
					break;

				case 3:	// DOWN
					xa = x;
					ya = y + 1;
					break;

				case 4:	// DOWN-RIGHT
					xa = x - 1;
					ya = y + x % 2;
					break;

				default:	// UP-RIGHT
					xa = x - 1;
					ya = y - 1 + x % 2;
					break;
			}		
		
			// Test if adjacent tile is part of board, and if so, add to list
			if (Valid(xa, ya)) a.add(d, xa, ya);
		}
		return a;	
	}
	
	// Retruns true iff a tile <x,y> is part of board
	private boolean Valid(int x, int y) {
		if ((x>=0 && x<sizeX) && (y>=0 && y<sizeY)) return true;
		return false;
	}
}
